@@ -1,96 +1,78 @@
+/* $Id: misc.c 3409 2006-01-19 17:50:40Z peter1138 $ */
+
 #include "stdafx.h"
-#include "ttd.h"
+#include "openttd.h"
+#include "currency.h"
+#include "functions.h"
+#include "news.h"
+#include "player.h"
+#include "string.h"
 #include "table/strings.h"
+#include "table/sprites.h"
 #include "map.h"
 #include "vehicle.h"
-#include "gfx.h"
-#include "assert.h"
 #include "saveload.h"
 #include "network.h"
 #include "network_data.h"
 #include "network_server.h"
+#include "engine.h"
+#include "vehicle_gui.h"
+#include "variables.h"
+#include "ai/ai.h"
 
-extern void StartupEconomy();
-extern void InitNewsItemStructs();
+extern void StartupEconomy(void);
 
-byte _name_array[512][32];
+char _name_array[512][32];
 
-static inline uint32 ROR(uint32 x, int n)
-{
-	return (x >> n) + (x << ((sizeof(x)*8)-n));
-}
-
-/* XXX - Player-seeds don't seem to be used anymore.. which is a good thing
-     so I just disabled them for now. If there are no problems, we can remove
-     it completely! -- TrueLight */
-#undef PLAYER_SEED_RANDOM
+#ifndef MERSENNE_TWISTER
 
 #ifdef RANDOM_DEBUG
 #include "network_data.h"
-
-uint32 DoRandom(uint line, char *file)
-#else
-uint32 Random()
-#endif
+uint32 DoRandom(int line, const char *file)
+#else // RANDOM_DEBUG
+uint32 Random(void)
+#endif // RANDOM_DEBUG
 {
+
+uint32 s;
+uint32 t;
+
 #ifdef RANDOM_DEBUG
 	if (_networking && (DEREF_CLIENT(0)->status != STATUS_INACTIVE || !_network_server))
 		printf("Random [%d/%d] %s:%d\n",_frame_counter, _current_player, file, line);
 #endif
 
-#ifdef PLAYER_SEED_RANDOM
-	if (_current_player>=MAX_PLAYERS || !_networking) {
-		uint32 s = _random_seeds[0][0];
-		uint32 t = _random_seeds[0][1];
-		_random_seeds[0][0] = s + ROR(t ^ 0x1234567F, 7);
-		return _random_seeds[0][1] = ROR(s, 3);
-	} else {
-		uint32 s = _player_seeds[_current_player][0];
-		uint32 t = _player_seeds[_current_player][1];
-		_player_seeds[_current_player][0] = s + ROR(t ^ 0x1234567F, 7);
-		DEBUG(net, 1)("[NET-Seeds] Player seed called!");
-		return _player_seeds[_current_player][1] = ROR(s, 3);
-	}
-#else
-	uint32 s = _random_seeds[0][0];
-	uint32 t = _random_seeds[0][1];
-	_random_seeds[0][0] = s + ROR(t ^ 0x1234567F, 7);
-	return _random_seeds[0][1] = ROR(s, 3);
-#endif
+	s = _random_seeds[0][0];
+	t = _random_seeds[0][1];
+	_random_seeds[0][0] = s + ROR(t ^ 0x1234567F, 7) + 1;
+	return _random_seeds[0][1] = ROR(s, 3) - 1;
 }
+#endif // MERSENNE_TWISTER
 
-#ifdef RANDOM_DEBUG
-uint DoRandomRange(uint max, uint line, char *file)
+#if defined(RANDOM_DEBUG) && !defined(MERSENNE_TWISTER)
+uint DoRandomRange(uint max, int line, const char *file)
 {
-	return (uint16)DoRandom(line, file) * max >> 16;
+	return GB(DoRandom(line, file), 0, 16) * max >> 16;
 }
 #else
 uint RandomRange(uint max)
 {
-	return (uint16)Random() * max >> 16;
+	return GB(Random(), 0, 16) * max >> 16;
 }
 #endif
 
-uint32 InteractiveRandom()
+
+uint32 InteractiveRandom(void)
 {
 	uint32 t = _random_seeds[1][1];
 	uint32 s = _random_seeds[1][0];
-	_random_seeds[1][0] = s + ROR(t ^ 0x1234567F, 7);
-	return _random_seeds[1][1] = ROR(s, 3);
+	_random_seeds[1][0] = s + ROR(t ^ 0x1234567F, 7) + 1;
+	return _random_seeds[1][1] = ROR(s, 3) - 1;
 }
 
 uint InteractiveRandomRange(uint max)
 {
-	return (uint16)InteractiveRandom() * max >> 16;
-}
-
-void InitPlayerRandoms()
-{
-	int i;
-	for (i=0; i<MAX_PLAYERS; i++) {
-		_player_seeds[i][0]=InteractiveRandom();
-		_player_seeds[i][1]=InteractiveRandom();
-	}
+	return GB(InteractiveRandom(), 0, 16) * max >> 16;
 }
 
 void SetDate(uint date)
@@ -100,105 +82,73 @@ void SetDate(uint date)
 	_cur_year = ymd.year;
 	_cur_month = ymd.month;
 #ifdef ENABLE_NETWORK
-	_network_last_advertise_date = 0;
+	_network_last_advertise_frame = 0;
+	_network_need_advertise = true;
 #endif /* ENABLE_NETWORK */
 }
 
+void InitializeVehicles(void);
+void InitializeWaypoints(void);
+void InitializeDepot(void);
+void InitializeEngines(void);
+void InitializeOrders(void);
+void InitializeClearLand(void);
+void InitializeRail(void);
+void InitializeRailGui(void);
+void InitializeRoad(void);
+void InitializeRoadGui(void);
+void InitializeAirportGui(void);
+void InitializeDock(void);
+void InitializeDockGui(void);
+void InitializeIndustries(void);
+void InitializeLandscape(void);
+void InitializeTowns(void);
+void InitializeTrees(void);
+void InitializeSigns(void);
+void InitializeStations(void);
+static void InitializeNameMgr(void);
+void InitializePlayers(void);
+static void InitializeCheats(void);
+void InitializeNPF(void);
 
-// multi os compatible sleep function
+void GenerateLandscape(void);
+void GenerateClearTile(void);
 
-#ifdef __AMIGA__
-// usleep() implementation
-#	include <devices/timer.h>
-#	include <dos/dos.h>
+void GenerateIndustries(void);
+void GenerateUnmovables(void);
+bool GenerateTowns(void);
 
-	extern struct Device      *TimerBase    = NULL;
-	extern struct MsgPort     *TimerPort    = NULL;
-	extern struct timerequest *TimerRequest = NULL;
-#endif // __AMIGA__
+void StartupPlayers(void);
+void StartupDisasters(void);
+void GenerateTrees(void);
 
-void CSleep(int milliseconds)
+void ConvertGroundTilesIntoWaterTiles(void);
+
+void InitializeGame(int mode, uint size_x, uint size_y)
 {
-	#if defined(WIN32)
-		Sleep(milliseconds);
-	#endif
-	#if defined(UNIX)
-		#if !defined(__BEOS__) && !defined(__AMIGA__)
-			usleep(milliseconds * 1000);
-		#endif
-		#ifdef __BEOS__
-			snooze(milliseconds * 1000);
-		#endif
-		#if defined(__AMIGA__)
-		{
-			ULONG signals;
-			ULONG TimerSigBit = 1 << TimerPort->mp_SigBit;
+	AllocateMap(size_x, size_y);
 
-			// send IORequest
-			TimerRequest->tr_node.io_Command = TR_ADDREQUEST;
-			TimerRequest->tr_time.tv_secs    = (milliseconds * 1000) / 1000000;
-			TimerRequest->tr_time.tv_micro   = (milliseconds * 1000) % 1000000;
-			SendIO((struct IORequest *)TimerRequest);
+	AddTypeToEngines(); // make sure all engines have a type
 
-			if (!((signals = Wait(TimerSigBit | SIGBREAKF_CTRL_C)) & TimerSigBit) ) {
-				AbortIO((struct IORequest *)TimerRequest);
-			}
-			WaitIO((struct IORequest *)TimerRequest);
-		}
-		#endif // __AMIGA__
-	#endif
-}
-
-void InitializeClearLand();
-void InitializeRail();
-void InitializeRailGui();
-void InitializeRoad();
-void InitializeRoadGui();
-void InitializeAirportGui();
-void InitializeDock();
-void InitializeDockGui();
-void InitializeIndustries();
-void InitializeLandscape();
-void InitializeTowns();
-void InitializeTrees();
-void InitializeStations();
-static void InitializeNameMgr();
-void InitializePlayers();
-static void InitializeCheats();
-
-void GenerateLandscape();
-void GenerateClearTile();
-
-void GenerateIndustries();
-void GenerateUnmovables();
-void GenerateTowns();
-
-void StartupPlayers();
-void StartupEngines();
-void StartupDisasters();
-void GenerateTrees();
-
-void ConvertGroundTilesIntoWaterTiles();
-
-void InitializeGame()
-{
-	SetObjectToPlace(1, 0, 0, 0);
+	SetObjectToPlace(SPR_CURSOR_ZZZ, 0, 0, 0);
 
 	_pause = 0;
 	_fast_forward = 0;
 	_tick_counter = 0;
 	_date_fract = 0;
 	_cur_tileloop_tile = 0;
-	_vehicle_id_ctr_day = 0;
 
-	{
+	if ((mode & IG_DATE_RESET) == IG_DATE_RESET) {
 		uint starting = ConvertIntDate(_patches.starting_date);
 		if ( starting == (uint)-1) starting = 10958;
 		SetDate(starting);
 	}
 
+	InitializeEngines();
 	InitializeVehicles();
-	_backup_orders_tile = 0;
+	InitializeWaypoints();
+	InitializeDepot();
+	InitializeOrders();
 
 	InitNewsItemStructs();
 	InitializeLandscape();
@@ -212,13 +162,16 @@ void InitializeGame()
 	InitializeDockGui();
 	InitializeTowns();
 	InitializeTrees();
+	InitializeSigns();
 	InitializeStations();
 	InitializeIndustries();
 
 	InitializeNameMgr();
 	InitializeVehiclesGuiList();
 	InitializeTrains();
+	InitializeNPF();
 
+	AI_Initialize();
 	InitializePlayers();
 	InitializeCheats();
 
@@ -231,7 +184,7 @@ void InitializeGame()
 	ResetObjectToPlace();
 }
 
-void GenerateWorld(int mode)
+void GenerateWorld(int mode, uint size_x, uint size_y)
 {
 	int i;
 
@@ -239,14 +192,14 @@ void GenerateWorld(int mode)
 	_current_player = OWNER_NONE;
 
 	_generating_world = true;
-	InitializeGame();
-	SetObjectToPlace(1, 0, 0, 0);
+	InitializeGame(mode == GW_RANDOM ? 0 : IG_DATE_RESET, size_x, size_y);
+	SetObjectToPlace(SPR_CURSOR_ZZZ, 0, 0, 0);
 
 	// Must start economy early because of the costs.
 	StartupEconomy();
 
 	// Don't generate landscape items when in the scenario editor.
-	if (mode == 1) {
+	if (mode == GW_EMPTY) {
 		// empty world in scenario editor
 		ConvertGroundTilesIntoWaterTiles();
 	} else {
@@ -254,7 +207,7 @@ void GenerateWorld(int mode)
 		GenerateClearTile();
 
 		// only generate towns, tree and industries in newgame mode.
-		if (mode == 0) {
+		if (mode == GW_NEWGAME) {
 			GenerateTowns();
 			GenerateTrees();
 			GenerateIndustries();
@@ -269,7 +222,7 @@ void GenerateWorld(int mode)
 	_generating_world = false;
 
 	// No need to run the tile loop in the scenario editor.
-	if (mode != 1) {
+	if (mode != GW_EMPTY) {
 		for(i=0x500; i!=0; i--)
 			RunTileLoop();
 	}
@@ -284,82 +237,46 @@ void DeleteName(StringID id)
 	}
 }
 
-byte *GetName(int id, byte *buff)
+char *GetName(int id, char *buff)
 {
-	byte *b;
-
-	if (id & 0x600) {
-		if (id & 0x200) {
-			if (id & 0x400) {
-				GetParamInt32();
-				GetParamUint16();
-			} else {
-				GetParamUint16();
-			}
-		} else {
-			GetParamInt32();
-		}
-	}
-
-	b = _name_array[(id&~0x600)];
-	while ((*buff++ = *b++) != 0);
-
-	return buff - 1;
+	return strecpy(buff, _name_array[id & ~0x600], NULL);
 }
 
 
-static void InitializeCheats()
+static void InitializeCheats(void)
 {
 	memset(&_cheats, 0, sizeof(Cheats));
 }
 
 
-static void InitializeNameMgr()
+static void InitializeNameMgr(void)
 {
 	memset(_name_array, 0, sizeof(_name_array));
 }
 
-StringID RealAllocateName(const byte *name, byte skip, bool check_double)
+StringID RealAllocateName(const char *name, byte skip, bool check_double)
 {
-	int free_item = -1;
-	const byte *names;
-	byte *dst;
-	int i;
+	char (*free_item)[lengthof(*_name_array)] = NULL;
+	char (*i)[lengthof(*_name_array)];
 
-	names = &_name_array[0][0];
-
-	for(i=0; i!=512; i++,names+=sizeof(_name_array[0])) {
-		if (names[0] == 0) {
-			if (free_item == -1)
-				free_item = i;
-		} else {
-			if (check_double && str_eq(names, name)) {
-				_error_message = STR_0132_CHOSEN_NAME_IN_USE_ALREADY;
-				return 0;
-			}
+	for (i = _name_array; i != endof(_name_array); ++i) {
+		if ((*i)[0] == '\0') {
+			if (free_item == NULL) free_item = i;
+		} else if (check_double && strncmp(*i, name, lengthof(*i) - 1) == 0) {
+			_error_message = STR_0132_CHOSEN_NAME_IN_USE_ALREADY;
+			return 0;
 		}
 	}
 
-	if (free_item < 0) {
+	if (free_item != NULL) {
+		ttd_strlcpy(*free_item, name, lengthof(*free_item));
+		return (free_item - _name_array) | 0x7800 | (skip << 8);
+	} else {
 		_error_message = STR_0131_TOO_MANY_NAMES_DEFINED;
 		return 0;
 	}
-
-	dst=_name_array[free_item];
-
-	for(i=0; (dst[i] = name[i]) != 0 && ++i != 32; ) {}
-	dst[31] = 0;
-
-	return free_item | 0x7800 | (skip << 8);
 }
 
-const TileIndexDiff _tileoffs_by_dir[4] = {
-	TILE_XY(-1, 0),
-	TILE_XY(0, 1),
-	TILE_XY(1, 0),
-	TILE_XY(0, -1),
-};
-
 
 #define M(a,b) ((a<<5)|b)
 static const uint16 _month_date_from_year_day[] = {
@@ -442,7 +359,7 @@ uint ConvertYMDToDay(uint year, uint month, uint day)
 }
 
 // convert a date on the form
-// 1920 - 2090
+// 1920 - 2090 (MAX_YEAR_END_REAL)
 // 192001 - 209012
 // 19200101 - 20901231
 // or if > 2090 and below 65536, treat it as a daycount
@@ -451,7 +368,7 @@ uint ConvertIntDate(uint date)
 {
 	uint year, month = 0, day = 1;
 
-	if (IS_INT_INSIDE(date, 1920, 2090 + 1)) {
+	if (IS_INT_INSIDE(date, 1920, MAX_YEAR_END_REAL + 1)) {
 		year = date - 1920;
 	} else if (IS_INT_INSIDE(date, 192001, 209012+1)) {
 		month = date % 100 - 1;
@@ -507,9 +424,7 @@ void InitializeLandscapeVariables(bool only_constants)
 
 		str = lpd->names[i];
 		_cargoc.names_s[i] = str;
-		_cargoc.names_p[i] = (str += 0x20);
-		_cargoc.names_long_s[i] = (str += 0x20);
-		_cargoc.names_long_p[i] = (str += 0x20);
+		_cargoc.names_long[i] = (str += 0x40);
 		_cargoc.names_short[i] = (str += 0x20);
 		_cargoc.weights[i] = lpd->weights[i];
 
@@ -523,49 +438,12 @@ void InitializeLandscapeVariables(bool only_constants)
 	}
 }
 
-// distance in Manhattan metric
-uint GetTileDist(TileIndex xy1, TileIndex xy2)
-{
-	return myabs(GET_TILE_X(xy1) - GET_TILE_X(xy2)) +
-				myabs(GET_TILE_Y(xy1) - GET_TILE_Y(xy2));
-}
-
-// maximum distance in x _or_ y
-uint GetTileDist1D(TileIndex xy1, TileIndex xy2)
-{
-	return max(myabs(GET_TILE_X(xy1) - GET_TILE_X(xy2)),
-						 myabs(GET_TILE_Y(xy1) - GET_TILE_Y(xy2)));
-}
-
-uint GetTileDist1Db(TileIndex xy1, TileIndex xy2)
-{
-	int a = myabs(GET_TILE_X(xy1) - GET_TILE_X(xy2));
-	int b = myabs(GET_TILE_Y(xy1) - GET_TILE_Y(xy2));
-
-	if (a > b)
-		return a*2+b;
-	else
-		return b*2+a;
-}
-
-uint GetTileDistAdv(TileIndex xy1, TileIndex xy2)
-{
-	uint a = myabs(GET_TILE_X(xy1) - GET_TILE_X(xy2));
-	uint b = myabs(GET_TILE_Y(xy1) - GET_TILE_Y(xy2));
-	return a*a+b*b;
-}
-
-bool CheckDistanceFromEdge(TileIndex tile, uint distance)
-{
-	return IS_INT_INSIDE(GET_TILE_X(tile), distance, TILE_X_MAX + 1 - distance) &&
-			IS_INT_INSIDE(GET_TILE_Y(tile), distance, TILE_Y_MAX + 1 - distance);
-}
 
 void OnNewDay_Train(Vehicle *v);
 void OnNewDay_RoadVeh(Vehicle *v);
 void OnNewDay_Aircraft(Vehicle *v);
 void OnNewDay_Ship(Vehicle *v);
-void OnNewDay_EffectVehicle(Vehicle *v) { /* empty */ }
+static void OnNewDay_EffectVehicle(Vehicle *v) { /* empty */ }
 void OnNewDay_DisasterVehicle(Vehicle *v);
 
 typedef void OnNewVehicleDayProc(Vehicle *v);
@@ -579,21 +457,21 @@ static OnNewVehicleDayProc * _on_new_vehicle_day_proc[] = {
 	OnNewDay_DisasterVehicle,
 };
 
-void EnginesDailyLoop();
-void DisasterDailyLoop();
-void PlayersMonthlyLoop();
-void EnginesMonthlyLoop();
-void TownsMonthlyLoop();
-void IndustryMonthlyLoop();
-void StationMonthlyLoop();
+void EnginesDailyLoop(void);
+void DisasterDailyLoop(void);
+void PlayersMonthlyLoop(void);
+void EnginesMonthlyLoop(void);
+void TownsMonthlyLoop(void);
+void IndustryMonthlyLoop(void);
+void StationMonthlyLoop(void);
 
-void PlayersYearlyLoop();
-void TrainsYearlyLoop();
-void RoadVehiclesYearlyLoop();
-void AircraftYearlyLoop();
-void ShipsYearlyLoop();
+void PlayersYearlyLoop(void);
+void TrainsYearlyLoop(void);
+void RoadVehiclesYearlyLoop(void);
+void AircraftYearlyLoop(void);
+void ShipsYearlyLoop(void);
 
-void WaypointsDailyLoop();
+void WaypointsDailyLoop(void);
 
 
 static const uint16 _autosave_months[] = {
@@ -604,9 +482,22 @@ static const uint16 _autosave_months[] = {
 	0x001, // every 12 months
 };
 
-void IncreaseDate()
+/**
+ * Runs the day_proc for every DAY_TICKS vehicle starting at daytick.
+ */
+static void RunVehicleDayProc(uint daytick)
+{
+	uint i, total = _vehicle_pool.total_items;
+
+	for (i = daytick; i < total; i += DAY_TICKS) {
+		Vehicle *v = GetVehicle(i);
+		if (v->type != 0)
+			_on_new_vehicle_day_proc[v->type - 0x10](v);
+	}
+}
+
+void IncreaseDate(void)
 {
-	int i,ctr,t;
 	YearMonthDay ymd;
 
 	if (_game_mode == GM_MENU) {
@@ -614,36 +505,19 @@ void IncreaseDate()
 		return;
 	}
 
-	/*if the day changed, call the vehicle event but only update a part of the vehicles
-		old max was i!= 12. But with that and a bigger number of vehicles (2560), per day only
-		a part of it could be done, namely: function called max_size date_fract (uint16) / 885 x 12 ==>
-		65536 / 885 = 74; 74x12 = 888. So max 888. Any vehicles above that were not _on_new_vehicle_day_proc'd
-		eg. aged.
-		So new code updates it for max vehicles.
-		(NUM_VEHICLES / maximum number of times ctr is incremented before reset ) + 1 (to get last vehicles too)
-		max size of _date_fract / 885 (added each tick) is number of times before ctr is reset.
-		Calculation might look complicated, but compiler just replaces it with 35, so that's ok
-	*/
-
-	ctr = _vehicle_id_ctr_day;
-	for(i=0; i!=(NUM_VEHICLES / ((1<<sizeof(_date_fract)*8) / 885)) + 1 && ctr != lengthof(_vehicles); i++) {
-		Vehicle *v = &_vehicles[ctr++];
-		if ((t=v->type) != 0)
-			_on_new_vehicle_day_proc[t - 0x10](v);
-	}
-	_vehicle_id_ctr_day = ctr;
+	RunVehicleDayProc(_date_fract);
 
 	/* increase day, and check if a new day is there? */
 	_tick_counter++;
 
-	if ( (_date_fract += 885) >= 885)
+	_date_fract++;
+	if (_date_fract < DAY_TICKS)
 		return;
+	_date_fract = 0;
 
 	/* yeah, increse day counter and call various daily loops */
 	_date++;
 
-	_vehicle_id_ctr_day = 0;
-
 	TextMessageDailyLoop();
 
 	DisasterDailyLoop();
@@ -695,14 +569,21 @@ void IncreaseDate()
 		NetworkServerYearlyLoop();
 #endif /* ENABLE_NETWORK */
 
-	/* check if we reached 2090, that's the maximum year. */
-	if (_cur_year == 171) {
+	/* check if we reached end of the game (31 dec 2050) */
+	if (_cur_year == _patches.ending_date - MAX_YEAR_BEGIN_REAL) {
+			ShowEndGameChart();
+	/* check if we reached 2090 (MAX_YEAR_END_REAL), that's the maximum year. */
+	} else if (_cur_year == (MAX_YEAR_END + 1)) {
 		Vehicle *v;
-		_cur_year = 170;
+		_cur_year = MAX_YEAR_END;
 		_date = 62093;
 		FOR_ALL_VEHICLES(v) {
 			v->date_of_last_service -= 365; // 1 year is 365 days long
 		}
+
+		/* Because the _date wraps here, and text-messages expire by game-days, we have to clean out
+		 *  all of them if the date is set back, else those messages will hang for ever */
+		InitTextMessage();
 	}
 
 	if (_patches.auto_euro)
@@ -713,10 +594,10 @@ void IncreaseDate()
 
 int FindFirstBit(uint32 value)
 {
-	// This is much faster then the one that was before here.
+	// This is much faster than the one that was before here.
 	//  Created by Darkvater.. blame him if it is wrong ;)
 	// Btw, the macro FINDFIRSTBIT is better to use when your value is
-	//  not more then 128.
+	//  not more than 128.
 	byte i = 0;
 	if (value & 0xffff0000) { value >>= 16; i += 16; }
 	if (value & 0x0000ff00) { value >>= 8;  i += 8; }
@@ -726,42 +607,49 @@ int FindFirstBit(uint32 value)
 	return i;
 }
 
+//!We're writing an own sort algorithm here, as
+//!qsort isn't stable
+//!Since the number of elements will be low, a
+//!simple bubble sort will have to do :)
 
-extern uint SafeTileAdd(uint tile, int add, const char *exp, const char *file, int line)
+void bubblesort(void *base, size_t nmemb, size_t size, int(*compar)(const void *, const void *))
 {
-	int x = GET_TILE_X(tile) + (signed char)(add & 0xFF);
-	int y = GET_TILE_Y(tile) + ((((0x8080 + add)>>8) & 0xFF) - 0x80);
+	uint i,k;
+	void *buffer = malloc(size);
+	char *start = base;
 
-	if (x < 0 || y < 0 || x >= TILES_X || y >= TILES_Y) {
-		char buf[512];
+	nmemb--;
 
-		sprintf(buf, "TILE_ADD(%s) when adding 0x%.4X and %d failed", exp, tile, add);
-#if !defined(_DEBUG) || !defined(_MSC_VER)
-		printf("%s\n", buf);
-#else
-		_assert(buf, (char*)file, line);
-#endif
+	for (i = 0; i < nmemb; i++) {
+		for (k = 0; k < nmemb; k++) {
+			void *a, *b;
+			a = start + size * k;
+			b = start + size * (k + 1);
+			if (compar(a, b) > 0) {
+				memcpy(buffer, a, size);
+				memcpy(a, b, size);
+				memcpy(b, buffer, size);
+			}
+		}
 	}
 
-	assert(TILE_XY(x,y) == TILE_MASK(tile + add));
-
-	return TILE_XY(x,y);
+	free(buffer);
+	buffer = NULL;
 }
 
-static void Save_NAME()
+static void Save_NAME(void)
 {
 	int i;
-	byte *b = _name_array[0];
 
-	for(i=0; i!=lengthof(_name_array); i++,b+=sizeof(_name_array[0])) {
-		if (*b) {
+	for (i = 0; i != lengthof(_name_array); ++i) {
+		if (_name_array[i][0] != '\0') {
 			SlSetArrayIndex(i);
-			SlArray(b, strlen(b), SLE_UINT8);
+			SlArray(_name_array[i], strlen(_name_array[i]), SLE_UINT8);
 		}
 	}
 }
 
-static void Load_NAME()
+static void Load_NAME(void)
 {
 	int index;
 
@@ -770,7 +658,7 @@ static void Load_NAME()
 	}
 }
 
-static const byte _game_opt_desc[] = {
+static const SaveLoad _game_opt_desc[] = {
 	// added a new difficulty option (town attitude) in version 4
 	SLE_CONDARR(GameOptions,diff,						SLE_FILE_I16 | SLE_VAR_I32, 17, 0, 3),
 	SLE_CONDARR(GameOptions,diff,						SLE_FILE_I16 | SLE_VAR_I32, 18, 4, 255),
@@ -786,7 +674,7 @@ static const byte _game_opt_desc[] = {
 };
 
 // Save load game options
-static void SaveLoad_OPTS()
+static void SaveLoad_OPTS(void)
 {
 	SlObject(&_opt, _game_opt_desc);
 }
@@ -799,69 +687,284 @@ static const SaveLoadGlobVarList _date_desc[] = {
 	{&_vehicle_id_ctr_day, 			SLE_UINT16, 0, 255},
 	{&_age_cargo_skip_counter, 	SLE_UINT8,	0, 255},
 	{&_avail_aircraft, 					SLE_UINT8,	0, 255},
-	{&_cur_tileloop_tile, 			SLE_UINT16, 0, 255},
+	{&_cur_tileloop_tile, 			SLE_FILE_U16 | SLE_VAR_U32, 0, 5},
+	{&_cur_tileloop_tile, 			SLE_UINT32, 6, 255},
 	{&_disaster_delay, 					SLE_UINT16, 0, 255},
 	{&_station_tick_ctr, 				SLE_UINT16, 0, 255},
 	{&_random_seeds[0][0], 					SLE_UINT32, 0, 255},
 	{&_random_seeds[0][1], 					SLE_UINT32, 0, 255},
-	{&_cur_town_ctr, 						SLE_UINT8,	0, 255},
+	{&_cur_town_ctr, 						SLE_FILE_U8 | SLE_VAR_U32,	0, 9},
+	{&_cur_town_ctr,						SLE_UINT32, 10, 255},
 	{&_cur_player_tick_index, 	SLE_FILE_U8 | SLE_VAR_UINT, 0, 255},
 	{&_next_competitor_start, 	SLE_FILE_U16 | SLE_VAR_UINT, 0, 255},
 	{&_trees_tick_ctr, 					SLE_UINT8,	0, 255},
 	{&_pause, 									SLE_UINT8,	4, 255},
+	{&_cur_town_iter, 						SLE_UINT32,	11, 255},
 	{NULL,											0,					0,   0}
 };
 
 // Save load date related variables as well as persistent tick counters
 // XXX: currently some unrelated stuff is just put here
-static void SaveLoad_DATE()
+static void SaveLoad_DATE(void)
 {
 	SlGlobList(_date_desc);
 }
 
 
 static const SaveLoadGlobVarList _view_desc[] = {
-	{&_saved_scrollpos_x,			SLE_FILE_I16 | SLE_VAR_INT, 0, 255},
-	{&_saved_scrollpos_y,			SLE_FILE_I16 | SLE_VAR_INT, 0, 255},
+	{&_saved_scrollpos_x,			SLE_FILE_I16 | SLE_VAR_INT, 0, 5},
+	{&_saved_scrollpos_x,			SLE_INT32, 6, 255},
+	{&_saved_scrollpos_y,			SLE_FILE_I16 | SLE_VAR_INT, 0, 5},
+	{&_saved_scrollpos_y,			SLE_INT32, 6, 255},
 	{&_saved_scrollpos_zoom,	SLE_UINT8,	0, 255},
 	{NULL,										0,					0,   0}
 };
 
-static void SaveLoad_VIEW()
+static void SaveLoad_VIEW(void)
 {
 	SlGlobList(_view_desc);
 }
 
-static void SaveLoad_MAPT() {
-  SlArray(_map_type_and_height, MapSize(), SLE_UINT8);
+static uint32 _map_dim_x;
+static uint32 _map_dim_y;
+
+static const SaveLoadGlobVarList _map_dimensions[] = {
+	{&_map_dim_x, SLE_UINT32, 6, 255},
+	{&_map_dim_y, SLE_UINT32, 6, 255},
+	{NULL, 0, 0, 0}
+};
+
+static void Save_MAPS(void)
+{
+	_map_dim_x = MapSizeX();
+	_map_dim_y = MapSizeY();
+	SlGlobList(_map_dimensions);
 }
 
-static void SaveLoad_MAP2() {
-  SlArray(_map2, MapSize(), SLE_UINT8);
+static void Load_MAPS(void)
+{
+	SlGlobList(_map_dimensions);
+	AllocateMap(_map_dim_x, _map_dim_y);
 }
 
-static void SaveLoad_M3LO() {
-  SlArray(_map3_lo, MapSize(), SLE_UINT8);
+static void Load_MAPT(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].type_height = buf[j];
+	}
 }
 
-static void SaveLoad_M3HI() {
-  SlArray(_map3_hi, MapSize(), SLE_UINT8);
+static void Save_MAPT(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size);
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].type_height;
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
 }
 
-static void SaveLoad_MAPO() {
-  SlArray(_map_owner, MapSize(), SLE_UINT8);
+static void Load_MAP1(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].m1 = buf[j];
+	}
 }
 
-static void SaveLoad_MAP5() {
-  SlArray(_map5, MapSize(), SLE_UINT8);
+static void Save_MAP1(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size);
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].m1;
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
 }
 
-static void SaveLoad_MAPE() {
-  SlArray(_map_extra_bits, MapSize() / 4, SLE_UINT8);
+static void Load_MAP2(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		uint16 buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf),
+			/* In those versions the m2 was 8 bits */
+			CheckSavegameVersion(5) ? SLE_FILE_U8 | SLE_VAR_U16 : SLE_UINT16
+		);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].m2 = buf[j];
+	}
+}
+
+static void Save_MAP2(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size * sizeof(_m[0].m2));
+	for (i = 0; i != size;) {
+		uint16 buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].m2;
+		SlArray(buf, lengthof(buf), SLE_UINT16);
+	}
+}
+
+static void Load_MAP3(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].m3 = buf[j];
+	}
+}
+
+static void Save_MAP3(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size);
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].m3;
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
+}
+
+static void Load_MAP4(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].m4 = buf[j];
+	}
+}
+
+static void Save_MAP4(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size);
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].m4;
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
+}
+
+static void Load_MAP5(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) _m[i++].m5 = buf[j];
+	}
+}
+
+static void Save_MAP5(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size);
+	for (i = 0; i != size;) {
+		byte buf[4096];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) buf[j] = _m[i++].m5;
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
+}
+
+static void Load_MAPE(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	for (i = 0; i != size;) {
+		uint8 buf[1024];
+		uint j;
+
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+		for (j = 0; j != lengthof(buf); j++) {
+			_m[i++].extra = GB(buf[j], 0, 2);
+			_m[i++].extra = GB(buf[j], 2, 2);
+			_m[i++].extra = GB(buf[j], 4, 2);
+			_m[i++].extra = GB(buf[j], 6, 2);
+		}
+	}
+}
+
+static void Save_MAPE(void)
+{
+	uint size = MapSize();
+	uint i;
+
+	SlSetLength(size / 4);
+	for (i = 0; i != size;) {
+		uint8 buf[1024];
+		uint j;
+
+		for (j = 0; j != lengthof(buf); j++) {
+			buf[j]  = _m[i++].extra << 0;
+			buf[j] |= _m[i++].extra << 2;
+			buf[j] |= _m[i++].extra << 4;
+			buf[j] |= _m[i++].extra << 6;
+		}
+		SlArray(buf, lengthof(buf), SLE_UINT8);
+	}
 }
 
 
-static void Save_CHTS()
+static void Save_CHTS(void)
 {
 	byte count = sizeof(_cheats)/sizeof(Cheat);
 	Cheat* cht = (Cheat*) &_cheats;
@@ -874,7 +977,7 @@ static void Save_CHTS()
 	}
 }
 
-static void Load_CHTS()
+static void Load_CHTS(void)
 {
 	Cheat* cht = (Cheat*) &_cheats;
 
@@ -888,13 +991,14 @@ static void Load_CHTS()
 
 
 const ChunkHandler _misc_chunk_handlers[] = {
-	{ 'MAPT', SaveLoad_MAPT, SaveLoad_MAPT, CH_RIFF },
-	{ 'MAP2', SaveLoad_MAP2, SaveLoad_MAP2, CH_RIFF },
-	{ 'M3LO', SaveLoad_M3LO, SaveLoad_M3LO, CH_RIFF },
-	{ 'M3HI', SaveLoad_M3HI, SaveLoad_M3HI, CH_RIFF },
-	{ 'MAPO', SaveLoad_MAPO, SaveLoad_MAPO, CH_RIFF },
-	{ 'MAP5', SaveLoad_MAP5, SaveLoad_MAP5, CH_RIFF },
-	{ 'MAPE', SaveLoad_MAPE, SaveLoad_MAPE, CH_RIFF },
+	{ 'MAPS', Save_MAPS, Load_MAPS, CH_RIFF },
+	{ 'MAPT', Save_MAPT, Load_MAPT, CH_RIFF },
+	{ 'MAPO', Save_MAP1, Load_MAP1, CH_RIFF },
+	{ 'MAP2', Save_MAP2, Load_MAP2, CH_RIFF },
+	{ 'M3LO', Save_MAP3, Load_MAP3, CH_RIFF },
+	{ 'M3HI', Save_MAP4, Load_MAP4, CH_RIFF },
+	{ 'MAP5', Save_MAP5, Load_MAP5, CH_RIFF },
+	{ 'MAPE', Save_MAPE, Load_MAPE, CH_RIFF },
 
 	{ 'NAME', Save_NAME, Load_NAME, CH_ARRAY},
 	{ 'DATE', SaveLoad_DATE, SaveLoad_DATE, CH_RIFF},
