@@ -1,111 +1,111 @@
+/* $Id: ship_cmd.c 3472 2006-01-29 18:57:26Z Darkvater $ */
+
 #include "stdafx.h"
-#include "ttd.h"
+#include "openttd.h"
 #include "table/strings.h"
+#include "functions.h"
 #include "map.h"
+#include "tile.h"
 #include "vehicle.h"
 #include "command.h"
 #include "pathfind.h"
 #include "station.h"
-#include "gfx.h"
 #include "news.h"
 #include "engine.h"
-#include "gui.h"
 #include "player.h"
 #include "sound.h"
+#include "npf.h"
+#include "depot.h"
+#include "vehicle_gui.h"
 
 static const uint16 _ship_sprites[] = {0x0E5D, 0x0E55, 0x0E65, 0x0E6D};
 static const byte _ship_sometracks[4] = {0x19, 0x16, 0x25, 0x2A};
 
-static byte GetTileShipTrackStatus(uint tile) {
+static byte GetTileShipTrackStatus(TileIndex tile)
+{
 	uint32 r = GetTileTrackStatus(tile, TRANSPORT_WATER);
 	return r | r >> 8;
 }
 
-void DrawShipEngine(int x, int y, int engine, uint32 image_ormod)
+void DrawShipEngine(int x, int y, EngineID engine, uint32 image_ormod)
 {
 	int spritenum = ShipVehInfo(engine)->image_index;
 
 	if (is_custom_sprite(spritenum)) {
 		int sprite = GetCustomVehicleIcon(engine, 6);
 
-		if (sprite) {
+		if (sprite != 0) {
 			DrawSprite(sprite | image_ormod, x, y);
 			return;
 		}
-		spritenum = _engine_original_sprites[engine];
+		spritenum = orig_ship_vehicle_info[engine - SHIP_ENGINES_INDEX].image_index;
 	}
 	DrawSprite((6 + _ship_sprites[spritenum]) | image_ormod, x, y);
 }
 
-void DrawShipEngineInfo(int engine, int x, int y, int maxw)
-{
-	const ShipVehicleInfo *svi = ShipVehInfo(engine);
-	SetDParam(0, svi->base_cost * (_price.ship_base>>3)>>5);
-	SetDParam(1, svi->max_speed * 10 >> 5);
-	SetDParam(2, _cargoc.names_long_p[svi->cargo_type]);
-	SetDParam(3, svi->capacity);
-	SetDParam(4, svi->running_cost * _price.ship_running >> 8);
-	DrawStringMultiCenter(x, y, STR_982E_COST_MAX_SPEED_CAPACITY, maxw);
-}
-
-int GetShipImage(Vehicle *v, byte direction)
+int GetShipImage(const Vehicle *v, byte direction)
 {
 	int spritenum = v->spritenum;
 
 	if (is_custom_sprite(spritenum)) {
 		int sprite = GetCustomVehicleSprite(v, direction);
 
-		if (sprite) return sprite;
-		spritenum = _engine_original_sprites[v->engine_type];
+		if (sprite != 0) return sprite;
+		spritenum = orig_ship_vehicle_info[v->engine_type - SHIP_ENGINES_INDEX].image_index;
 	}
 	return _ship_sprites[spritenum] + direction;
 }
 
-static int FindClosestShipDepot(Vehicle *v)
+static const Depot* FindClosestShipDepot(const Vehicle* v)
 {
-	uint tile, dist, best_dist = (uint)-1;
-	int best_depot = -1;
-	int i;
-	byte owner = v->owner;
-	uint tile2 = v->tile;
+	const Depot* depot;
+	const Depot* best_depot = NULL;
+	uint dist;
+	uint best_dist = (uint)-1;
+	TileIndex tile;
+	TileIndex tile2 = v->tile;
 
-	for(i=0; i!=lengthof(_depots); i++) {
-		tile = _depots[i].xy;
-		if (IS_TILETYPE(tile, MP_WATER) && _map_owner[tile] == owner) {
-			dist = GetTileDist(tile, tile2);
-			if (dist < best_dist) {
-				best_dist = dist;
-				best_depot = i;
+	if (_patches.new_pathfinding_all) {
+		NPFFoundTargetData ftd;
+		byte trackdir = GetVehicleTrackdir(v);
+		ftd = NPFRouteToDepotTrialError(v->tile, trackdir, TRANSPORT_WATER, v->owner, INVALID_RAILTYPE);
+		if (ftd.best_bird_dist == 0) {
+			best_depot = GetDepotByTile(ftd.node.tile); /* Found target */
+		} else {
+			best_depot = NULL; /* Did not find target */
+		}
+	} else {
+		FOR_ALL_DEPOTS(depot) {
+			tile = depot->xy;
+			if (IsValidDepot(depot) && IsTileDepotType(tile, TRANSPORT_WATER) && IsTileOwner(tile, v->owner)) {
+				dist = DistanceManhattan(tile, tile2);
+				if (dist < best_dist) {
+					best_dist = dist;
+					best_depot = depot;
+				}
 			}
 		}
 	}
-
 	return best_depot;
 }
 
 static void CheckIfShipNeedsService(Vehicle *v)
 {
-	int i;
+	const Depot* depot;
 
-	if (_patches.servint_ships == 0)
-		return;
-
-	if (!VehicleNeedsService(v))
-		return;
-
-	if (v->vehstatus & VS_STOPPED)
-		return;
+	if (_patches.servint_ships == 0) return;
+	if (!VehicleNeedsService(v))     return;
+	if (v->vehstatus & VS_STOPPED)   return;
 
 	if (v->current_order.type == OT_GOTO_DEPOT &&
-			v->current_order.flags & OF_FULL_LOAD)
+			v->current_order.flags & OF_HALT_IN_DEPOT)
 		return;
 
-	if (_patches.gotodepot && ScheduleHasDepotOrders(v->schedule_ptr))
-		return;
+	if (_patches.gotodepot && VehicleHasDepotOrders(v)) return;
 
-	i = FindClosestShipDepot(v);
+	depot = FindClosestShipDepot(v);
 
-	if (i < 0 || GetTileDist(v->tile, (&_depots[i])->xy) > 12) {
+	if (depot == NULL || DistanceManhattan(v->tile, depot->xy) > 12) {
 		if (v->current_order.type == OT_GOTO_DEPOT) {
 			v->current_order.type = OT_DUMMY;
 			v->current_order.flags = 0;
@@ -116,8 +116,8 @@ static void CheckIfShipNeedsService(Vehicle *v)
 
 	v->current_order.type = OT_GOTO_DEPOT;
 	v->current_order.flags = OF_NON_STOP;
-	v->current_order.station = (byte)i;
-	v->dest_tile = (&_depots[i])->xy;
+	v->current_order.station = depot->index;
+	v->dest_tile = depot->xy;
 	InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
 }
 
@@ -132,12 +132,9 @@ void OnNewDay_Ship(Vehicle *v)
 	AgeVehicle(v);
 	CheckIfShipNeedsService(v);
 
-	CheckOrders(v);
-
-	if (v->vehstatus & VS_STOPPED)
-		return;
-
+	CheckOrders(v->index, OC_INIT);
 
+	if (v->vehstatus & VS_STOPPED) return;
 
 	cost = ShipVehInfo(v->engine_type)->running_cost * _price.ship_running / 364;
 	v->profit_this_year -= cost >> 8;
@@ -146,7 +143,8 @@ void OnNewDay_Ship(Vehicle *v)
 	SubtractMoneyFromPlayerFract(v->owner, cost);
 
 	InvalidateWindow(WC_VEHICLE_DETAILS, v->index);
-	InvalidateWindow(WC_SHIPS_LIST, v->owner);
+	//we need this for the profit
+	InvalidateWindowClasses(WC_SHIPS_LIST);
 }
 
 static void HandleBrokenShip(Vehicle *v)
@@ -166,8 +164,7 @@ static void HandleBrokenShip(Vehicle *v)
 
 		if (!(v->vehstatus & VS_HIDDEN)) {
 			Vehicle *u = CreateEffectVehicleRel(v, 4, 4, 5, EV_BREAKDOWN_SMOKE);
-			if (u)
-				u->u.special.unk0 = v->breakdown_delay * 2;
+			if (u != NULL) u->u.special.unk0 = v->breakdown_delay * 2;
 		}
 	}
 
@@ -190,21 +187,20 @@ static void PlayShipSound(Vehicle *v)
 	SndPlayVehicleFx(ShipVehInfo(v->engine_type)->sfx, v);
 }
 
-static const TileIndexDiff _dock_offs[] = {
-	TILE_XY(2, 0),
-	TILE_XY(-2, 0),
-	TILE_XY(0, 2),
-	TILE_XY(2, 0),
-	TILE_XY(0, -2),
-	TILE_XY(0,0),
-	TILE_XY(0,0),
-	TILE_XY(0,0),
+static const TileIndexDiffC _dock_offs[] = {
+	{ 2,  0},
+	{-2,  0},
+	{ 0,  2},
+	{ 2,  0},
+	{ 0, -2},
+	{ 0,  0},
+	{ 0,  0},
+	{ 0,  0}
 };
 
 static void ProcessShipOrder(Vehicle *v)
 {
-	Order order;
-	Station *st;
+	const Order *order;
 
 	if (v->current_order.type >= OT_GOTO_DEPOT &&
 			v->current_order.type <= OT_LEAVESTATION) {
@@ -214,7 +210,7 @@ static void ProcessShipOrder(Vehicle *v)
 	}
 
 	if (v->current_order.type == OT_GOTO_DEPOT &&
-			(v->current_order.flags & (OF_UNLOAD | OF_FULL_LOAD)) == (OF_UNLOAD | OF_FULL_LOAD) &&
+			(v->current_order.flags & (OF_PART_OF_ORDERS | OF_SERVICE_IF_NEEDED)) == (OF_PART_OF_ORDERS | OF_SERVICE_IF_NEEDED) &&
 			!VehicleNeedsService(v)) {
 		v->cur_order_index++;
 	}
@@ -223,49 +219,50 @@ static void ProcessShipOrder(Vehicle *v)
 	if (v->cur_order_index >= v->num_orders)
 		v->cur_order_index = 0;
 
-	order = v->schedule_ptr[v->cur_order_index];
+	order = GetVehicleOrder(v, v->cur_order_index);
 
-	if (order.type == OT_NOTHING) {
-		v->current_order.type = OT_NOTHING;
+	if (order == NULL) {
+		v->current_order.type  = OT_NOTHING;
 		v->current_order.flags = 0;
 		v->dest_tile = 0;
 		return;
 	}
 
-	if (order.type == v->current_order.type &&
-			order.flags == v->current_order.flags &&
-			order.station == v->current_order.station)
+	if (order->type    == v->current_order.type &&
+			order->flags   == v->current_order.flags &&
+			order->station == v->current_order.station)
 		return;
 
-	v->current_order = order;
+	v->current_order = *order;
 
-	if (order.type == OT_GOTO_STATION) {
-		if (order.station == v->last_station_visited)
-			v->last_station_visited = 0xFF;
+	if (order->type == OT_GOTO_STATION) {
+		const Station *st;
 
-		st = DEREF_STATION(order.station);
+		if (order->station == v->last_station_visited)
+			v->last_station_visited = INVALID_STATION;
+
+		st = GetStation(order->station);
 		if (st->dock_tile != 0) {
-			v->dest_tile = TILE_ADD(st->dock_tile, _dock_offs[_map5[st->dock_tile]-0x4B]);
+			v->dest_tile = TILE_ADD(st->dock_tile, ToTileIndexDiff(_dock_offs[_m[st->dock_tile].m5-0x4B]));
 		}
-	} else if (order.type == OT_GOTO_DEPOT) {
-		v->dest_tile = _depots[order.station].xy;
+	} else if (order->type == OT_GOTO_DEPOT) {
+		v->dest_tile = GetDepot(order->station)->xy;
 	} else {
 		v->dest_tile = 0;
 	}
-	InvalidateVehicleOrderWidget(v);
+
+	InvalidateVehicleOrder(v);
+
+	InvalidateWindowClasses(WC_SHIPS_LIST);
 }
 
 static void HandleShipLoading(Vehicle *v)
 {
-	if (v->current_order.type == OT_NOTHING)
-		return;
+	if (v->current_order.type == OT_NOTHING) return;
 
 	if (v->current_order.type != OT_DUMMY) {
-		if (v->current_order.type != OT_LOADING)
-			return;
-
-		if (--v->load_unload_time_rem)
-			return;
+		if (v->current_order.type != OT_LOADING) return;
+		if (--v->load_unload_time_rem) return;
 
 		if (v->current_order.flags & OF_FULL_LOAD && CanFillVehicle(v)) {
 			SET_EXPENSES_TYPE(EXPENSES_SHIP_INC);
@@ -281,13 +278,12 @@ static void HandleShipLoading(Vehicle *v)
 			Order b = v->current_order;
 			v->current_order.type = OT_LEAVESTATION;
 			v->current_order.flags = 0;
-			if (!(b.flags & OF_NON_STOP))
-				return;
+			if (!(b.flags & OF_NON_STOP)) return;
 		}
 	}
 
 	v->cur_order_index++;
-	InvalidateVehicleOrderWidget(v);
+	InvalidateVehicleOrder(v);
 }
 
 static void UpdateShipDeltaXY(Vehicle *v, int dir)
@@ -305,10 +301,10 @@ static void UpdateShipDeltaXY(Vehicle *v, int dir)
 	};
 #undef MKIT
 	uint32 x = _delta_xy_table[dir];
-	v->x_offs = (byte)x;
-	v->y_offs = (byte)(x>>=8);
-	v->sprite_width = (byte)(x>>=8);
-	v->sprite_height = (byte)(x>>=8);
+	v->x_offs        = GB(x,  0, 8);
+	v->y_offs        = GB(x,  8, 8);
+	v->sprite_width  = GB(x, 16, 8);
+	v->sprite_height = GB(x, 24, 8);
 }
 
 static void RecalcShipStuff(Vehicle *v)
@@ -319,34 +315,33 @@ static void RecalcShipStuff(Vehicle *v)
 	InvalidateWindow(WC_VEHICLE_DEPOT, v->tile);
 }
 
-static const TileIndexDiff _ship_leave_depot_offs[] = {
-	TILE_XY(-1,0),
-	TILE_XY(0,-1),
+static const TileIndexDiffC _ship_leave_depot_offs[] = {
+	{-1,  0},
+	{ 0, -1}
 };
 
 static void CheckShipLeaveDepot(Vehicle *v)
 {
-	uint tile;
+	TileIndex tile;
 	int d;
 	uint m;
 
-	if (v->u.ship.state != 0x80)
-		return;
+	if (v->u.ship.state != 0x80) return;
 
 	tile = v->tile;
-	d = (_map5[tile]&2) ? 1 : 0;
+	d = (_m[tile].m5&2) ? 1 : 0;
 
 	// Check first side
-	if (_ship_sometracks[d] & GetTileShipTrackStatus(TILE_ADD(tile,_ship_leave_depot_offs[d]))) {
+	if (_ship_sometracks[d] & GetTileShipTrackStatus(TILE_ADD(tile, ToTileIndexDiff(_ship_leave_depot_offs[d])))) {
 		m = (d==0) ? 0x101 : 0x207;
 	// Check second side
-	} else if (_ship_sometracks[d+2] & GetTileShipTrackStatus(TILE_ADD(tile, -2*_ship_leave_depot_offs[d]))) {
+	} else if (_ship_sometracks[d+2] & GetTileShipTrackStatus(TILE_ADD(tile, -2 * ToTileIndexDiff(_ship_leave_depot_offs[d])))) {
 		m = (d==0) ? 0x105 : 0x203;
 	} else {
 		return;
 	}
-	v->direction = (byte)m;
-	v->u.ship.state = (byte)(m >> 8);
+	v->direction    = GB(m, 0, 8);
+	v->u.ship.state = GB(m, 8, 8);
 	v->vehstatus &= ~VS_HIDDEN;
 
 	v->cur_speed = 0;
@@ -354,6 +349,7 @@ static void CheckShipLeaveDepot(Vehicle *v)
 
 	PlayShipSound(v);
 	VehicleServiceInDepot(v);
+	InvalidateWindowClasses(WC_SHIPS_LIST);
 }
 
 static bool ShipAccelerate(Vehicle *v)
@@ -371,23 +367,20 @@ static bool ShipAccelerate(Vehicle *v)
 	}
 
 	// Decrease somewhat when turning
-	if (!(v->direction&1)) {
-		spd = spd * 3 >> 2;
-	}
+	if (!(v->direction & 1)) spd = spd * 3 / 4;
 
-	if (spd == 0)
-		return false;
-
-	if ((byte)++spd == 0)
-		return true;
+	if (spd == 0) return false;
+	if ((byte)++spd == 0) return true;
 
 	v->progress = (t = v->progress) - (byte)spd;
 
 	return (t < v->progress);
 }
 
-
-static int32 EstimateShipCost(uint16 engine_type);
+static int32 EstimateShipCost(EngineID engine_type)
+{
+	return ShipVehInfo(engine_type)->base_cost * (_price.ship_base>>3)>>5;
+}
 
 static void ShipEnterDepot(Vehicle *v)
 {
@@ -400,8 +393,6 @@ static void ShipEnterDepot(Vehicle *v)
 
 	InvalidateWindow(WC_VEHICLE_DETAILS, v->index);
 
-	MaybeRenewVehicle(v, EstimateShipCost(v->engine_type));
-
 	TriggerVehicle(v, VEHICLE_TRIGGER_DEPOT);
 
 	if (v->current_order.type == OT_GOTO_DEPOT) {
@@ -413,9 +404,9 @@ static void ShipEnterDepot(Vehicle *v)
 		v->current_order.type = OT_DUMMY;
 		v->current_order.flags = 0;
 
-		if (t.flags & OF_UNLOAD) {
+		if (HASBIT(t.flags, OFB_PART_OF_ORDERS)) {
 			v->cur_order_index++;
-		} else if (t.flags & OF_FULL_LOAD) {
+		} else if (HASBIT(t.flags, OFB_HALT_IN_DEPOT)) {
 			v->vehstatus |= VS_STOPPED;
 			if (v->owner == _local_player) {
 				SetDParam(0, v->unitnumber);
@@ -427,14 +418,17 @@ static void ShipEnterDepot(Vehicle *v)
 			}
 		}
 	}
+	InvalidateWindowClasses(WC_SHIPS_LIST);
 }
 
-static void ShipArrivesAt(Vehicle *v, Station *st)
+static void ShipArrivesAt(const Vehicle* v, Station* st)
 {
 	/* Check if station was ever visited before */
 	if (!(st->had_vehicle_of_type & HVOT_SHIP)) {
 		uint32 flags;
+
 		st->had_vehicle_of_type |= HVOT_SHIP;
+
 		SetDParam(0, st->index);
 		flags = (v->owner == _local_player) ? NEWS_FLAGS(NM_THIN, NF_VIEWPORT|NF_VEHICLE, NT_ARRIVAL_PLAYER, 0) : NEWS_FLAGS(NM_THIN, NF_VIEWPORT|NF_VEHICLE, NT_ARRIVAL_OTHER, 0);
 		AddNewsItem(
@@ -446,30 +440,25 @@ static void ShipArrivesAt(Vehicle *v, Station *st)
 }
 
 typedef struct {
-	uint skiptile;
-	uint dest_coords;
+	TileIndex skiptile;
+	TileIndex dest_coords;
 	uint best_bird_dist;
 	uint best_length;
 } PathFindShip;
 
-//extern void dbg_store_path();
-
-static bool ShipTrackFollower(uint tile, PathFindShip *pfs, int track, uint length, byte *state)
+static bool ShipTrackFollower(TileIndex tile, PathFindShip *pfs, int track, uint length, byte *state)
 {
 	// Found dest?
 	if (tile == pfs->dest_coords) {
 		pfs->best_bird_dist = 0;
 
-//		if (length < pfs->best_length)
-//			dbg_store_path();
-
 		pfs->best_length = minu(pfs->best_length, length);
 		return true;
 	}
 
 	// Skip this tile in the calculation
 	if (tile != pfs->skiptile) {
-		pfs->best_bird_dist = minu(pfs->best_bird_dist, GetTileDist1Db(pfs->dest_coords, tile));
+		pfs->best_bird_dist = minu(pfs->best_bird_dist, DistanceMaxPlusManhattan(pfs->dest_coords, tile));
 	}
 
 	return false;
@@ -486,7 +475,7 @@ static const byte _ship_search_directions[6][4] = {
 
 static const byte _pick_shiptrack_table[6] = {1, 3, 2, 2, 0, 0};
 
-static uint FindShipTrack(Vehicle *v, uint tile, int dir, uint bits, uint skiptile, int *track)
+static uint FindShipTrack(Vehicle *v, TileIndex tile, int dir, uint bits, TileIndex skiptile, int *track)
 {
 	PathFindShip pfs;
 	int i, best_track;
@@ -521,7 +510,7 @@ static uint FindShipTrack(Vehicle *v, uint tile, int dir, uint bits, uint skipti
 
 			/* if we reach this position, there's two paths of equal value so far.
 			 * pick one randomly. */
-			r = (byte)Random();
+			r = GB(Random(), 0, 8);
 			if (_pick_shiptrack_table[i] == ship_dir) r += 80;
 			if (_pick_shiptrack_table[best_track] == ship_dir) r -= 80;
 			if (r <= 127) goto bad;
@@ -538,27 +527,56 @@ bad:;
 	return best_bird_dist;
 }
 
-static int ChooseShipTrack(Vehicle *v, uint tile, int dir, uint tracks)
+/* returns the track to choose on the next tile, or -1 when it's better to
+ * reverse. The tile given is the tile we are about to enter, enterdir is the
+ * direction in which we are entering the tile */
+static int ChooseShipTrack(Vehicle *v, TileIndex tile, int enterdir, uint tracks)
 {
-	uint b;
-	uint tot_dist, dist;
-	int track;
-	uint tile2;
+	assert(enterdir>=0 && enterdir<=3);
 
-	assert(dir>=0 && dir<=3);
-	tile2 = TILE_ADD(tile, -_tileoffs_by_dir[dir]);
-	dir ^= 2;
-	tot_dist = (uint)-1;
-	b = GetTileShipTrackStatus(tile2) & _ship_sometracks[dir] & v->u.ship.state;
-	if (b != 0) {
-		dist = FindShipTrack(v, tile2, dir, b, tile, &track);
-		if (dist != (uint)-1)
-			tot_dist = dist + 1;
+	if (_patches.new_pathfinding_all) {
+		NPFFindStationOrTileData fstd;
+		NPFFoundTargetData ftd;
+		TileIndex src_tile = TILE_ADD(tile, TileOffsByDir(ReverseDiagdir(enterdir)));
+		byte trackdir = GetVehicleTrackdir(v);
+		assert (trackdir != 0xFF); /* Check that we are not in a depot */
+
+		NPFFillWithOrderData(&fstd, v);
+
+		ftd = NPFRouteToStationOrTile(src_tile, trackdir, &fstd, TRANSPORT_WATER, v->owner, INVALID_RAILTYPE);
+
+		if (ftd.best_trackdir != 0xff) {
+			/* If ftd.best_bird_dist is 0, we found our target and ftd.best_trackdir contains
+			the direction we need to take to get there, if ftd.best_bird_dist is not 0,
+			we did not find our target, but ftd.best_trackdir contains the direction leading
+			to the tile closest to our target. */
+			return ftd.best_trackdir & 7; /* TODO: Wrapper function? */
+		} else {
+			return -1; /* Already at target, reverse? */
+		}
+	} else {
+		uint b;
+		uint tot_dist, dist;
+		int track;
+		TileIndex tile2;
+
+		tile2 = TILE_ADD(tile, -TileOffsByDir(enterdir));
+		tot_dist = (uint)-1;
+
+		/* Let's find out how far it would be if we would reverse first */
+		b = GetTileShipTrackStatus(tile2) & _ship_sometracks[ReverseDiagdir(enterdir)] & v->u.ship.state;
+		if (b != 0) {
+			dist = FindShipTrack(v, tile2, ReverseDiagdir(enterdir), b, tile, &track);
+			if (dist != (uint)-1)
+				tot_dist = dist + 1;
+		}
+		/* And if we would not reverse? */
+		dist = FindShipTrack(v, tile, enterdir, tracks, 0, &track);
+		if (dist > tot_dist)
+			/* We could better reverse */
+			return -1;
+		return track;
 	}
-	dist = FindShipTrack(v, tile, dir^2, tracks, 0, &track);
-	if (dist > tot_dist)
-		return -1;
-	return track;
 }
 
 static const byte _new_vehicle_direction_table[11] = {
@@ -567,10 +585,10 @@ static const byte _new_vehicle_direction_table[11] = {
 	2, 3, 4,
 };
 
-static int ShipGetNewDirectionFromTiles(uint new_tile, uint old_tile)
+static int ShipGetNewDirectionFromTiles(TileIndex new_tile, TileIndex old_tile)
 {
-	uint offs = (GET_TILE_Y(new_tile) - GET_TILE_Y(old_tile) + 1) * 4 +
-							GET_TILE_X(new_tile) - GET_TILE_X(old_tile) + 1;
+	uint offs = (TileY(new_tile) - TileY(old_tile) + 1) * 4 +
+							TileX(new_tile) - TileX(old_tile) + 1;
 	assert(offs < 11 && offs != 3 && offs != 7);
 	return _new_vehicle_direction_table[offs];
 }
@@ -582,7 +600,7 @@ static int ShipGetNewDirection(Vehicle *v, int x, int y)
 	return _new_vehicle_direction_table[offs];
 }
 
-static int GetAvailShipTracks(uint tile, int dir)
+static int GetAvailShipTracks(TileIndex tile, int dir)
 {
 	uint32 r = GetTileTrackStatus(tile, TRANSPORT_WATER);
 	return (byte) ((r | r >> 8)) & _ship_sometracks[dir];
@@ -623,7 +641,6 @@ static const byte _ship_subcoord[4][6][3] = {
 	}
 };
 
-
 static void ShipController(Vehicle *v)
 {
 	GetNewVehiclePosResult gp;
@@ -641,19 +658,16 @@ static void ShipController(Vehicle *v)
 		v->breakdown_ctr--;
 	}
 
-	if (v->vehstatus & VS_STOPPED)
-		return;
+	if (v->vehstatus & VS_STOPPED) return;
 
 	ProcessShipOrder(v);
 	HandleShipLoading(v);
 
-	if (v->current_order.type == OT_LOADING)
-		return;
+	if (v->current_order.type == OT_LOADING) return;
 
 	CheckShipLeaveDepot(v);
 
-	if (!ShipAccelerate(v))
-		return;
+	if (!ShipAccelerate(v)) return;
 
 	BeginVehicleMove(v);
 
@@ -667,53 +681,64 @@ static void ShipController(Vehicle *v)
 			r = VehicleEnterTile(v, gp.new_tile, gp.x, gp.y);
 			if (r & 0x8) goto reverse_direction;
 
-			if (v->dest_tile != 0 && v->dest_tile == gp.new_tile) {
-				if (v->current_order.type == OT_GOTO_DEPOT) {
-					if ((gp.x&0xF)==8 && (gp.y&0xF)==8) {
-						ShipEnterDepot(v);
-						return;
-					}
-				} else if (v->current_order.type == OT_GOTO_STATION) {
-					Station *st;
-
-					v->last_station_visited = v->current_order.station;
-
-					/* Process station in the schedule. Don't do that for buoys (HVOT_BUOY) */
-					st = DEREF_STATION(v->current_order.station);
-					if (!(st->had_vehicle_of_type & HVOT_BUOY) 
-							&& (st->facilities & FACIL_DOCK)) { /* ugly, ugly workaround for problem with ships able to drop off cargo at wrong stations */
-						v->current_order.type = OT_LOADING;
-						v->current_order.flags &= OF_FULL_LOAD | OF_UNLOAD;
-						v->current_order.flags |= OF_NON_STOP;
-						ShipArrivesAt(v, st);
-
-						SET_EXPENSES_TYPE(EXPENSES_SHIP_INC);
-						if (LoadUnloadVehicle(v)) {
-							InvalidateWindow(WC_SHIPS_LIST, v->owner);
-							MarkShipDirty(v);
-						}
-						InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
-					} else { /* leave buoys right aways */
-						v->current_order.type = OT_LEAVESTATION;
-						v->current_order.flags = 0;
-						v->cur_order_index++;
-						InvalidateVehicleOrderWidget(v);
-					}
-					goto else_end;
-				}
-			}
-
+			/* A leave station order only needs one tick to get processed, so we can
+			 * always skip ahead. */
 			if (v->current_order.type == OT_LEAVESTATION) {
 				v->current_order.type = OT_NOTHING;
 				v->current_order.flags = 0;
 				InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
+			} else if (v->dest_tile != 0) {
+				/* We have a target, let's see if we reached it... */
+				if (v->current_order.type == OT_GOTO_STATION &&
+						IsBuoyTile(v->dest_tile) &&
+						DistanceManhattan(v->dest_tile, gp.new_tile) <= 3) {
+					/* We got within 3 tiles of our target buoy, so let's skip to our
+					 * next order */
+					v->cur_order_index++;
+					v->current_order.type = OT_DUMMY;
+					InvalidateVehicleOrder(v);
+				} else {
+					/* Non-buoy orders really need to reach the tile */
+					if (v->dest_tile == gp.new_tile) {
+						if (v->current_order.type == OT_GOTO_DEPOT) {
+							if ((gp.x&0xF)==8 && (gp.y&0xF)==8) {
+								ShipEnterDepot(v);
+								return;
+							}
+						} else if (v->current_order.type == OT_GOTO_STATION) {
+							Station *st;
+
+							v->last_station_visited = v->current_order.station;
+
+							/* Process station in the orderlist. */
+							st = GetStation(v->current_order.station);
+							if (st->facilities & FACIL_DOCK) { /* ugly, ugly workaround for problem with ships able to drop off cargo at wrong stations */
+								v->current_order.type = OT_LOADING;
+								v->current_order.flags &= OF_FULL_LOAD | OF_UNLOAD;
+								v->current_order.flags |= OF_NON_STOP;
+								ShipArrivesAt(v, st);
+
+								SET_EXPENSES_TYPE(EXPENSES_SHIP_INC);
+								if (LoadUnloadVehicle(v)) {
+									InvalidateWindow(WC_SHIPS_LIST, v->owner);
+									MarkShipDirty(v);
+								}
+								InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
+							} else { /* leave stations without docks right aways */
+								v->current_order.type = OT_LEAVESTATION;
+								v->current_order.flags = 0;
+								v->cur_order_index++;
+								InvalidateVehicleOrder(v);
+							}
+						}
+					}
+				}
 			}
 		}
 	} else {
 		// new tile
-		if (GET_TILE_X(gp.new_tile) == 0xFF ||
-				(byte)GET_TILE_Y(gp.new_tile) == 0xFF)
-					goto reverse_direction;
+		if (TileX(gp.new_tile) >= MapMaxX() || TileY(gp.new_tile) >= MapMaxY())
+			goto reverse_direction;
 
 		dir = ShipGetNewDirectionFromTiles(gp.new_tile, gp.old_tile);
 		assert(dir == 1 || dir == 3 || dir == 5 || dir == 7);
@@ -743,7 +768,6 @@ static void ShipController(Vehicle *v)
 
 		v->direction = b[2];
 	}
-else_end:;
 
 	/* update image of ship, as well as delta XY */
 	dir = ShipGetNewDirection(v, gp.x, gp.y);
@@ -766,10 +790,8 @@ reverse_direction:
 
 static void AgeShipCargo(Vehicle *v)
 {
-	if (_age_cargo_skip_counter != 0)
-		return;
-	if (v->cargo_days != 255)
-		v->cargo_days++;
+	if (_age_cargo_skip_counter != 0) return;
+	if (v->cargo_days != 255) v->cargo_days++;
 }
 
 void Ship_Tick(Vehicle *v)
@@ -778,12 +800,8 @@ void Ship_Tick(Vehicle *v)
 	ShipController(v);
 }
 
-void HandleClickOnShip(Vehicle *v)
-{
-	ShowShipViewWindow(v);
-}
 
-void ShipsYearlyLoop()
+void ShipsYearlyLoop(void)
 {
 	Vehicle *v;
 
@@ -796,28 +814,33 @@ void ShipsYearlyLoop()
 	}
 }
 
-static int32 EstimateShipCost(uint16 engine_type)
-{
-	return ShipVehInfo(engine_type)->base_cost * (_price.ship_base>>3)>>5;
-}
-
-// p1 = type to build
+/** Build a ship.
+ * @param x,y tile coordinates of depot where ship is built
+ * @param p1 ship type being built (engine)
+ * @param p2 unused
+ */
 int32 CmdBuildShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	int32 value;
 	Vehicle *v;
-	uint unit_num;
-	uint tile = TILE_FROM_XY(x,y);
+	UnitID unit_num;
+	TileIndex tile = TileVirtXY(x, y);
 	Engine *e;
 
+	if (!IsEngineBuildable(p1, VEH_Ship)) return_cmd_error(STR_ENGINE_NOT_BUILDABLE);
+
 	SET_EXPENSES_TYPE(EXPENSES_NEW_VEHICLES);
 
 	value = EstimateShipCost(p1);
-	if (flags & DC_QUERY_COST)
-		return value;
+	if (flags & DC_QUERY_COST) return value;
+
+	/* The ai_new queries the vehicle cost before building the route,
+	 * so we must check against cheaters no sooner than now. --pasky */
+	if (!IsTileDepotType(tile, TRANSPORT_WATER)) return CMD_ERROR;
+	if (!IsTileOwner(tile, _current_player)) return CMD_ERROR;
 
 	v = AllocateVehicle();
-	if (v == NULL || _ptr_to_next_order >= endof(_order_array) ||
+	if (v == NULL || IsOrderPoolFull() ||
 			(unit_num = GetFreeUnitNumber(VEH_Ship)) > _patches.max_ships)
 		return_cmd_error(STR_00E1_TOO_MANY_VEHICLES_IN_GAME);
 
@@ -828,8 +851,8 @@ int32 CmdBuildShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 
 		v->owner = _current_player;
 		v->tile = tile;
-		x = GET_TILE_X(tile)*16 + 8;
-		y = GET_TILE_Y(tile)*16 + 8;
+		x = TileX(tile) * 16 + 8;
+		y = TileY(tile) * 16 + 8;
 		v->x_pos = x;
 		v->y_pos = y;
 		v->z_pos = GetSlopeZ(x,y);
@@ -846,50 +869,57 @@ int32 CmdBuildShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 		v->cargo_cap = svi->capacity;
 		v->value = value;
 
-		v->last_station_visited = 255;
+		v->last_station_visited = INVALID_STATION;
 		v->max_speed = svi->max_speed;
-		v->engine_type = (byte)p1;
+		v->engine_type = p1;
 
-		e = &_engines[p1];
+		e = GetEngine(p1);
 		v->reliability = e->reliability;
 		v->reliability_spd_dec = e->reliability_spd_dec;
 		v->max_age = e->lifelength * 366;
 		_new_ship_id = v->index;
+		_new_vehicle_id = v->index;
 
 		v->string_id = STR_SV_SHIP_NAME;
 		v->u.ship.state = 0x80;
-		_ptr_to_next_order->type = OT_NOTHING;
-		_ptr_to_next_order->flags = 0;
-		v->schedule_ptr = _ptr_to_next_order++;
 
 		v->service_interval = _patches.servint_ships;
 		v->date_of_last_service = _date;
 		v->build_year = _cur_year;
 		v->cur_image = 0x0E5E;
 		v->type = VEH_Ship;
+		v->random_bits = VehicleRandomBits();
 
 		VehiclePositionChanged(v);
 
 		InvalidateWindow(WC_VEHICLE_DEPOT, v->tile);
 		RebuildVehicleLists();
 		InvalidateWindow(WC_COMPANY, v->owner);
+		if (IsLocalPlayer())
+			InvalidateWindow(WC_REPLACE_VEHICLE, VEH_Ship); // updates the replace Ship window
 	}
 
 	return value;
 }
 
+/** Sell a ship.
+ * @param x,y unused
+ * @param p1 vehicle ID to be sold
+ * @param p2 unused
+ */
 int32 CmdSellShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	Vehicle *v;
 
+	if (!IsVehicleIndex(p1)) return CMD_ERROR;
+
+	v = GetVehicle(p1);
+
+	if (v->type != VEH_Ship || !CheckOwnership(v->owner)) return CMD_ERROR;
+
 	SET_EXPENSES_TYPE(EXPENSES_NEW_VEHICLES);
 
-	v = &_vehicles[p1];
-
-	if (!CheckOwnership(v->owner))
-		return CMD_ERROR;
-
-	if (!IsShipDepotTile(v->tile) || v->u.road.state != 0x80 || !(v->vehstatus&VS_STOPPED))
+	if (!IsTileDepotType(v->tile, TRANSPORT_WATER) || v->u.road.state != 0x80 || !(v->vehstatus&VS_STOPPED))
 		return_cmd_error(STR_980B_SHIP_MUST_BE_STOPPED_IN);
 
 	if (flags & DC_EXEC) {
@@ -898,113 +928,126 @@ int32 CmdSellShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 		InvalidateWindow(WC_COMPANY, v->owner);
 		DeleteWindowById(WC_VEHICLE_VIEW, v->index);
 		DeleteVehicle(v);
+		if (IsLocalPlayer())
+			InvalidateWindow(WC_REPLACE_VEHICLE, VEH_Ship); // updates the replace Ship window
 	}
 
 	return -(int32)v->value;
 }
 
-// p1 = vehicle
+/** Start/Stop a ship.
+ * @param x,y unused
+ * @param p1 ship ID to start/stop
+ * @param p2 unused
+ */
 int32 CmdStartStopShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	Vehicle *v;
 
-	v = &_vehicles[p1];
+	if (!IsVehicleIndex(p1)) return CMD_ERROR;
 
-	if (!CheckOwnership(v->owner))
-		return CMD_ERROR;
+	v = GetVehicle(p1);
+
+	if (v->type != VEH_Ship || !CheckOwnership(v->owner)) return CMD_ERROR;
 
 	if (flags & DC_EXEC) {
 		v->vehstatus ^= VS_STOPPED;
 		InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
 		InvalidateWindow(WC_VEHICLE_DEPOT, v->tile);
+		InvalidateWindowClasses(WC_SHIPS_LIST);
 	}
 
 	return 0;
 }
 
+/** Send a ship to the depot.
+ * @param x,y unused
+ * @param p1 vehicle ID to send to the depot
+ * @param p2 unused
+ */
 int32 CmdSendShipToDepot(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	Vehicle *v;
-	int depot;
+	const Depot *dep;
 
-	v = &_vehicles[p1];
+	if (!IsVehicleIndex(p1)) return CMD_ERROR;
 
-	if (!CheckOwnership(v->owner))
-		return CMD_ERROR;
+	v = GetVehicle(p1);
 
+	if (v->type != VEH_Ship || !CheckOwnership(v->owner)) return CMD_ERROR;
+
+	if (v->vehstatus & VS_CRASHED) return CMD_ERROR;
+
+	/* If the current orders are already goto-depot */
 	if (v->current_order.type == OT_GOTO_DEPOT) {
 		if (flags & DC_EXEC) {
-			if (v->current_order.flags & OF_UNLOAD) v->cur_order_index++;
+			/* If the orders to 'goto depot' are in the orders list (forced servicing),
+			 * then skip to the next order; effectively cancelling this forced service */
+			if (HASBIT(v->current_order.flags, OFB_PART_OF_ORDERS))
+				v->cur_order_index++;
+
 			v->current_order.type = OT_DUMMY;
 			v->current_order.flags = 0;
 			InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
 		}
-	} else {
-		depot = FindClosestShipDepot(v);
-		if (depot < 0)
-			return_cmd_error(STR_981A_UNABLE_TO_FIND_LOCAL_DEPOT);
-
-		if (flags & DC_EXEC) {
-			v->dest_tile = _depots[depot].xy;
-			v->current_order.type = OT_GOTO_DEPOT;
-			v->current_order.flags = OF_NON_STOP | OF_FULL_LOAD;
-			v->current_order.station = depot;
-			InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
-		}
+		return 0;
 	}
 
-	return 0;
-}
-
-int32 CmdChangeShipServiceInt(int x, int y, uint32 flags, uint32 p1, uint32 p2)
-{
-	Vehicle *v;
-
-	v = &_vehicles[p1];
-
-	if (!CheckOwnership(v->owner))
-		return CMD_ERROR;
+	dep = FindClosestShipDepot(v);
+	if (dep == NULL)
+		return_cmd_error(STR_981A_UNABLE_TO_FIND_LOCAL_DEPOT);
 
 	if (flags & DC_EXEC) {
-		v->service_interval = (uint16)p2;
-		InvalidateWindowWidget(WC_VEHICLE_DETAILS, v->index, 7);
+		v->dest_tile = dep->xy;
+		v->current_order.type = OT_GOTO_DEPOT;
+		v->current_order.flags = OF_NON_STOP | OF_HALT_IN_DEPOT;
+		v->current_order.station = dep->index;
+		InvalidateWindowWidget(WC_VEHICLE_VIEW, v->index, STATUS_BAR);
 	}
 
 	return 0;
 }
 
 
-// p1 = vehicle
-// p2 = new cargo
+/** Refits a ship to the specified cargo type.
+ * @param x,y unused
+ * @param p1 vehicle ID of the ship to refit
+ * @param p2 various bitstuffed elements
+ * - p2 = (bit 0-7) - the new cargo type to refit to (p2 & 0xFF)
+ */
 int32 CmdRefitShip(int x, int y, uint32 flags, uint32 p1, uint32 p2)
 {
 	Vehicle *v;
 	int32 cost;
+	CargoID new_cid = p2 & 0xFF; //gets the cargo number
+
+	if (!IsVehicleIndex(p1)) return CMD_ERROR;
+
+	v = GetVehicle(p1);
+
+	if (v->type != VEH_Ship || !CheckOwnership(v->owner)) return CMD_ERROR;
+
+	if (!IsTileDepotType(v->tile, TRANSPORT_WATER) || !(v->vehstatus&VS_STOPPED) || v->u.ship.state != 0x80)
+			return_cmd_error(STR_980B_SHIP_MUST_BE_STOPPED_IN);
+
+
+	/* Check cargo */
+	if (!ShipVehInfo(v->engine_type)->refittable) return CMD_ERROR;
+	if (new_cid > NUM_CARGO || !CanRefitTo(v->engine_type, new_cid)) return CMD_ERROR;
 
 	SET_EXPENSES_TYPE(EXPENSES_SHIP_RUN);
 
-	v = &_vehicles[p1];
-	if (!CheckOwnership(v->owner))
-		return CMD_ERROR;
-
-	if (!IsShipDepotTile(v->tile) ||
-			!(v->vehstatus&VS_STOPPED) ||
-			v->u.ship.state != 0x80)
-		return_cmd_error(STR_980B_SHIP_MUST_BE_STOPPED_IN);
-
 	cost = 0;
-	if (IS_HUMAN_PLAYER(v->owner) && (byte)p2 != v->cargo_type) {
+	if (IS_HUMAN_PLAYER(v->owner) && new_cid != v->cargo_type) {
 		cost = _price.ship_base >> 7;
 	}
 
 	if (flags & DC_EXEC) {
 		v->cargo_count = 0;
-		v->cargo_type = (byte)p2;
+		v->cargo_type = new_cid;
 		InvalidateWindow(WC_VEHICLE_DETAILS, v->index);
 	}
 
 	return cost;
 
 }
-
-
